/// Josh's User Interface Library
/// A C++20 Library for creating, styling, and rendering of a UI/UX widgets.
/// Developed and Maintained by Josh O'Leary @ Redacted Software.
/// Special Thanks to William Tomasine II and Maxine Hayes.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file Widget.hpp
/// @desc Base Widget Class - All JUI widgets extend their behavior from this class.
/// @edit 2024-07-31

#pragma once

#include <Event.h>
#include <EventConnection.h>
#include <string>
#include <vector>
#include "J3ML/LinearAlgebra.hpp"
#include "JUI/UDim2.hpp"
#include <Color3.hpp>
#include <Color4.hpp>

#include <JGL/JGL.h>
using namespace JGL;

namespace JUI
{

    enum class MouseButton {
        Left = 1,
        Middle = 2,
        Right = 3
    };

    using namespace J3ML::Math;
    using namespace J3ML::LinearAlgebra;

    /// Widget class is the base for all JUI elements.
    /// Widgets exist in a tree hierarchy where each object has one parent, and an arbitrary number of children.
    class Widget
    {
    public:
        /// The default constructor for Widget. Generally, JUI widgets will initialize member data,
        /// and apply a default style in this constructor. Be aware of this, as this is not "idiomatic C++".
        Widget();
        /// Construct a new widget by specifying it's parent,
        Widget(Widget* parent);
        virtual ~Widget() {}
    public:
        /// An event that triggers when the widget is in-focus, generally when the mouse is hovering it.
        /// (The actual trigger may differ for each widget)
        Event<> Focused;
        /// An event that triggers when the widget loses focus.
        Event<> Unfocused;
        /// This event triggers when a new widget is added to this widget, or any descendant of this widget.
        Event<Widget *> DescendantAdded;
        /// This event triggers when a widget is removed from this widget's children, or any of it's childrens' children, and so forth.
        Event<Widget *> DescendantRemoved;
        /// This event triggers when the hierarchy this widget is contained within changes.
        Event<Widget *, Widget*> AncestryChanged;
        /// This event triggers when a widget is added to this widget's list of children.
        Event<Widget *> ChildAdded;
        /// This event triggers when a widget is removed from this widget's list of children.
        Event<Widget *> ChildRemoved;
        /// This event triggers right before this widget gets deallocated.
        Event<Widget *> Destroying;
    public:


        /// Adds a given widget to this widget's list of children.
        /// @return The widget in question.
        Widget* Add(Widget* newChild);
        bool IsAncestorOf(Widget* descendant) const;
        bool IsDescendantOf(Widget* ancestor) const;

        /// Returns the first child widget that has the given symbolic name, Otherwise, a nullpointer is returned
        /// TODO: Use std::optional here and anywhere else a nullptr could be returned.
        Widget* FindFirstChild(const std::string& name);

        /// Returns a flat list of all nodes that are lower in the hierarchy list.
        std::vector<Widget*> GetDescendants();
        /// Returns a flat list of all nodes that are higher in the hierarchy list.
        std::vector<Widget*> GetAncestors();
        /// Returns the nodes directly descendant to this widget.
        std::vector<Widget*> GetChildren();

        /// Calculates the final size in pixels of this element.
        [[nodiscard]] virtual Vector2 GetAbsoluteSize() const;
        /// Calculates the final position in pixels of this element.
        [[nodiscard]] virtual Vector2 GetAbsolutePosition() const;
        [[nodiscard]] virtual float GetAbsoluteRotation() const;

        /// Returns the parent of this widget
        /// TODO: Use std::optional here and anywhere else a nullptr could be returned.
        Widget* GetParent() const;

        /// Returns the menu-coordinates that are used to position this widget in relation to its parent.
        /// @see class UDim2, Position(const UDim2&),
        [[nodiscard]] UDim2 Position() const;

        /// Retrieves this widgets z-index. This value determines the order in which a widget
        /// renders to the screen relative to other Widgets.
        /// @note This applies in ascending priority order,
        /// meaning lower values are rendered first, placing higher values on top of lower values.
        /// @note The range of valid values is -MAX_INT to MAX_INT.
        /// @note This does not manipulate the OpenGL Z buffer, rather, when rendering,
        /// we sort the direct children of a widget widget and render.
        int ZIndex() const;

        /// Sets this widgets z-index.
        /// @see ZIndex().
        void SetZIndex(int zindex);

        /// Returns the menu-coordinates that are used to size this widget in relation to its parent.
        /// @see SetSize(), class UDim2
        [[nodiscard]] UDim2 Size() const;

        /// Sets the widgets pixel and scalar position using a combined data type.
        /// This position is centered around the object's anchor point.
        /// @see Position(), SetAnchorPoint(), GetAnchorPoint(), class UDim2.
        void SetPosition(const UDim2 &pos);

        /// Sets this widget's pixel and scalar size using a combined data type.
        /// @see Size(), class UDim2.
        void SetSize(const UDim2 &s);

        /// Sets the parent object of this widget. Positioning and sizing of a widget is relative to it's parent.
        void SetParent(Widget*);

        /// Returns true if this widget is a 'descendant' of the specified potential ancestor. Otherwise returns false.
        bool IsDescendantOf(Widget *ancestor);

        /// Returns true if this widget is a 'ancestor' of the specified potential descendant. Otherwise returns false.
        bool IsAncestorOf(Widget *descendant);

        /// Determines the origin point of a Widget, relative to it's absolute size.
        [[nodiscard]] Vector2 GetAnchorPoint() const;

        /// Returns the padding factor on the left of this widget.
        /// @see SetPaddingLeft(), class UDim.
        [[nodiscard]] UDim GetPaddingLeft() const;
        /// Returns the padding factor on the top of this widget.
        /// @see SetPaddingTop(), class UDim.
        [[nodiscard]] UDim GetPaddingTop() const;
        /// Returns the padding factor on the right of this widget.
        /// @see SetPaddingRight(), class UDim.
        [[nodiscard]] UDim GetPaddingRight() const;
        /// Returns the padding factor on the bottom of this widget.
        /// @see SetPaddingBottom(), class UDim.
        [[nodiscard]] UDim GetPaddingBottom() const;

        void SetPaddingLeft(const UDim &pad_left);
        void SetPaddingTop(const UDim &pad_top);
        void SetPaddingRight(const UDim &pad_right);
        void SetPaddingBottom(const UDim &pad_bottom);

        void SetPadding(const UDim& left, const UDim& top, const UDim& right, const UDim& bottom);

        void SetPadding(const UDim& padding);

        /// Returns the margin factor on the left of this widget.
        [[nodiscard]] UDim GetMarginLeft() const;
        /// Returns the margin factor on the top of this widget.
        [[nodiscard]] UDim GetMarginTop() const;
        /// Returns the margin factor on the right of this widget.
        [[nodiscard]] UDim GetMarginRight() const;

        /// Returns the margin factor on the bottom of this widget.
        [[nodiscard]] UDim GetMarginBottom() const;

        /// Sets the amount (in Pixels + Scale) to apply margin on the left-side of the widget.
        /// @see UDim, SetMargin()
        void SetMarginLeft(const UDim &ml);
        /// Sets the amount (in Pixels + Scale) to apply margin on the top-side of the widget.
        /// @see UDim, SetMargin()
        void SetMarginTop(const UDim &mt);
        /// Sets the amount (in Pixels + Scale) to apply margin on the right-side of the widget.
        /// @see UDim, SetMargin()
        void SetMarginRight(const UDim &mr);
        /// Sets the amount (in Pixels + Scale) to apply margin on the bottom-side of the widget.
        /// @see UDim, SetMargin()
        void SetMarginBottom(const UDim &mb);

        /// Sets the margin factor of each side of the widget.
        /// @param left The amount of margin to apply on the left. (Pixels, Scale)
        /// @param top The amount of margin to apply on the top.
        /// @param right The amount of margin to apply on the right.
        /// @param bottom The amount of margin to apply on the bottom.
        void SetMargin(const UDim& left, const UDim& top, const UDim& right, const UDim& bottom);

        /// Sets the margin factor on all four sides of this widget to the same value.
        /// @see SetMargin(const UDim&, const UDim&, const UDim&, const UDim&).
        void SetMargin(const UDim& margin);

        /// Returns this widgets mnemonic name.
        /// Widgets can optionally be assigned a name that can be used to retrieve it from a widget tree node.
        /// @see GetName().
        [[nodiscard]] std::string GetName() const;

        /// Sets this widgets mnemonic name.
        /// Widgets can optionally be assigned a name that can be used to retrieve it from a widget tree node.
        /// @see GetName().
        void SetName(const std::string &new_name);

        /// Returns whether the widget is to be rendered.
        /// This function does not indicate whether the widget is **actually seeable** on-screen.
        /// @see SetVisible().
        [[nodiscard]] bool IsVisible() const;

        /// Sets whether the widget is to be rendered. Children are also not rendered if disabled.
        /// @see IsVisible().
        void SetVisible(bool enabled);


        /// Returns the first ancestor in this widgets hierarchy that does not have its own parent.
        /// In a well-formed JUI menu, this **should** always be a Scene.
        Widget* GetFamilyTreeRoot() const;

        bool IsMouseInside() const;

        void SetAnchorPoint(const Vector2 &point);

    public:
        /// Renders the widget to the current OpenGL Context using JGL.
        /// The user should call this on their Scene instances only. JUI will handle the rest.
        virtual void Draw();
        /// Performs the update logic of the widget.
        /// The user should call this on their Scene instances only. JUI will handle the rest.
        virtual void Update(float delta);
        /// Informs a widget that the mouse has been moved to a new location.
        /// This is designed in such a way that the end-user can plug this into their existing code.
        /// The user should call this on their Scene instances only. JUI will handle the rest.
        /// See ReWindowIntegrationDemo for an example.
        virtual void ObserveMouseMovement(const Vector2& latest_known_pos);
        /// Informs a widget that a mouse button has pressed or released.
        /// This is designed in such a way that the end-user can plug this into their existing code.
        /// The user should call this on their Scene instances only. JUI will handle the rest.
        /// See ReWindowIntegrationDemo for an example.
        virtual void ObserveMouseInput(MouseButton btn, bool pressed);
    protected:
        void DrawChildWidgets();
        void UpdateChildWidgets(float delta);
    protected:
        MouseButton mbtn;
        bool mb_state;
        bool prev_mb_state;
        //int last_known_mouse_button;
        //bool last_known_mouse_button_state;
        Vector2 last_known_mouse_pos;
        UDim2 position;
        UDim2 size;
        Widget* parent = nullptr;
        std::vector<Widget*> children;
        float rotation = 0;
        std::string name;
        bool selected = false;
        UDim pad_left = 0_px;
        UDim pad_right = 0_px;
        UDim pad_top = 0_px;
        UDim pad_bottom = 0_px;
        UDim margin_left = 0_px;
        UDim margin_right = 0_px;
        UDim margin_top = 0_px;
        UDim margin_bottom = 0_px;
        Vector2 anchor_point = {0.f, 0.f};
        bool visible = true;
        Widget* next = nullptr;
        Widget* prev = nullptr;


    };

    /// A mixin helper class that provides behavior for hoverable objects.
    /// A hoverable object pays attention to when the mouse enters and leaves it's bounds.
    class Hoverable {
    public:
        Event<Vector2> OnHoverEvent;
        Event<Vector2> OnExitEvent;
    public:
        bool IsHovered() const { return hovered; };
    public:
        virtual void OnHover(const Vector2& MousePos) {
            OnHoverEvent.Invoke(MousePos);
        };
        virtual void OnExit(const Vector2& MousePos) {
            OnExitEvent.Invoke(MousePos);
        };

        void Update(const Vector2& m_pos, float delta)
        {
            if (IsHovered() && !hover_debounce) {
                OnHover(m_pos);
                hover_debounce = true;
            }

            if (!IsHovered() && hover_debounce) {
                OnExit(m_pos);
                hover_debounce = false;
            }
        }
    protected:
        bool hovered;
        bool hover_debounce;
    };

    /// A mixin helper class that provides behavior and events for clickable widgets.
    class Clickable
    {
    public:
        Event<Vector2, MouseButton> OnClickEvent;
        Event<Vector2, MouseButton, bool> OnReleaseEvent;
    public:
        bool IsClicked() const { return clicked; };
        void SetClicked(bool manual_click)
        {
            clicked = manual_click;
        }
    public:
        virtual void OnClick(const Vector2& MousePos, const MouseButton& MouseButton) {
            OnClickEvent.Invoke(MousePos, MouseButton);
        };
        virtual void OnRelease(const Vector2& MousePos, const MouseButton& MouseButton, bool MouseStillOver) {
            OnReleaseEvent.Invoke(MousePos, MouseButton, MouseStillOver);
        };


        void Update(const Vector2& mpos, const MouseButton& btn, bool hovering)
        {

        }
    protected:
        bool clicked = false;
        bool click_debounce = false;
    };

    /// A mixin helper class that provides behavior and events for a binary-state togglable widget.
    class Toggleable {
    public:
        Event<> OnToggleEvent;
        Event<> OnToggleOnEvent;
        Event<> OnToggleOffEvent;
    public:
        virtual void OnToggleOn() {
            OnToggleOnEvent.Invoke();
        }
        virtual void OnToggleOff() {
            OnToggleOffEvent.Invoke();
        }
        virtual void Update() {
            OnToggleEvent.Invoke();
            toggled = !toggled;

            if (toggled) {
                OnToggleOn();
            } else {
                OnToggleOff();
            }
        }
    protected:
        bool toggled = false;
    };
}
